# This file is part of cloud-init. See LICENSE file for license information.
"""Datasource for Oracle (OCI/Oracle Cloud Infrastructure)

OCI provides a OpenStack like metadata service which provides only
'2013-10-17' and 'latest' versions..

Notes:
 * This datasource does not support the OCI-Classic. OCI-Classic
   provides an EC2 lookalike metadata service.
 * The uuid provided in DMI data is not the same as the meta-data provided
   instance-id, but has an equivalent lifespan.
 * We do need to support upgrade from an instance that cloud-init
   identified as OpenStack.
 * Both bare-metal and vms use iscsi root
 * Both bare-metal and vms provide chassis-asset-tag of OracleCloud.com
"""

from cloudinit.url_helper import combine_url, readurl, UrlError
from cloudinit.net import dhcp, get_interfaces_by_mac
from cloudinit.event import EventType
from cloudinit import net
from cloudinit import sources
from cloudinit import util
from cloudinit.net import cmdline
from cloudinit import log as logging

import json
import re

LOG = logging.getLogger(__name__)

DEFAULT_MTU = 9000
BUILTIN_DS_CONFIG = {
    'configure_secondary_nics': False,
    'mtu': DEFAULT_MTU,
}

CHASSIS_ASSET_TAG = "OracleCloud.com"
METADATA_ENDPOINT = "http://169.254.169.254/openstack/"
VNIC_METADATA_ENDPOINT = "http://169.254.169.254/opc/v1/vnics/"

class DataSourceOracle(sources.DataSource):

    dsname = 'Oracle'
    system_uuid = None
    vendordata_pure = None
    network_config_sources = (
        sources.NetworkConfigSource.cmdline,
        sources.NetworkConfigSource.ds,
        sources.NetworkConfigSource.initramfs,
        sources.NetworkConfigSource.system_cfg,
    )

    _network_config = sources.UNSET
    _vnic_metadata = sources.UNSET

    def __init__(self, sys_cfg, *args, **kwargs):
        super(DataSourceOracle, self).__init__(sys_cfg, *args, **kwargs)

        self.ds_cfg = util.mergemanydict([
            util.get_cfg_by_path(sys_cfg, ['datasource', self.dsname], {}),
            BUILTIN_DS_CONFIG])


    def _is_platform_viable(self):
        """Check platform environment to report if this datasource may run."""
        return _is_platform_viable()

    def _get_data(self):
        if not self._is_platform_viable():
            return False

        # network may be configured if iscsi root.  If that is the case
        # then read_initramfs_config will return non-None.
        if _is_iscsi_root():
            data, vnic_md = self.crawl_metadata()
        else:
            with dhcp.EphemeralDHCPv4(net.find_fallback_nic()):
                data, vnic_md = self.crawl_metadata()

        self._crawled_metadata = data
        vdata = data['2013-10-17']

        self.userdata_raw = vdata.get('user_data')
        self.system_uuid = vdata['system_uuid']

        vd = vdata.get('vendor_data')
        if vd:
            self.vendordata_pure = vd
            try:
                self.vendordata_raw = sources.convert_vendordata(vd)
            except ValueError as e:
                LOG.warning("Invalid content in vendor-data: %s", e)
                self.vendordata_raw = None

        mdcopies = ('public_keys',)
        md = dict([(k, vdata['meta_data'].get(k))
                   for k in mdcopies if k in vdata['meta_data']])

        mdtrans = (
            # oracle meta_data.json name, cloudinit.datasource.metadata name
            ('availability_zone', 'availability-zone'),
            ('hostname', 'local-hostname'),
            ('launch_index', 'launch-index'),
            ('uuid', 'instance-id'),
        )
        for dsname, ciname in mdtrans:
            if dsname in vdata['meta_data']:
                md[ciname] = vdata['meta_data'][dsname]

        self.metadata = md
        self._vnic_metadata = vnic_md

        self.ds_cfg = util.mergemanydict([
            util.get_cfg_by_path(self.sys_cfg, ['datasource', self.dsname], {}),
            BUILTIN_DS_CONFIG])

        return True

    def crawl_metadata(self):
        return read_metadata(), read_vnic_metadata()

    def _get_subplatform(self):
        """Return the subplatform metadata source details."""
        return 'metadata (%s)' % METADATA_ENDPOINT

    def check_instance_id(self, sys_cfg):
        """quickly check (local only) if self.instance_id is still valid

        On Oracle, the dmi-provided system uuid differs from the instance-id
        but has the same life-span."""
        return sources.instance_id_matches_system_uuid(self.system_uuid)

    def get_public_ssh_keys(self):
        return sources.normalize_pubkey_data(self.metadata.get('public_keys'))

    @property
    def update_events(self):
        events = {'network': set([EventType.BOOT_NEW_INSTANCE,
                                  EventType.BOOT]),
                  'storage': set([])}

        self.ds_cfg = util.mergemanydict([
            util.get_cfg_by_path(self.sys_cfg, ['datasource', self.dsname], {}),
            BUILTIN_DS_CONFIG])

        if self.ds_cfg.get('configure_secondary_nics', False):
            events['network'].add(EventType.HOTPLUG)

        LOG.debug('Oracle ds update events: %s', events)

        return events

    @property
    def network_config(self):
        """Network config is read from initramfs provided files, plus
        those got from Instance MetaData Service for secondary NICs.
        If none is present, then we fall back to fallback configuration.

        One thing to note here is that this method is not currently
        considered at all if there is kernel bootline provided
        data.  In that case, stages considers that the cmdline data
        overrides datasource provided data and does not consult here.

        The primary NIC is always configured through initramfs provided
        data. If initramfs provided data is unavailable, attempt will
        still be made to get metadata from IMDS for secondary NICs, once
        network for primary NIC is brought up by system. If no network
        is availabe at all, it has to generate a fallback config."""

        if self._network_config != sources.UNSET and bool(self._network_config):
            return self._network_config

        initrd_cfg = cmdline.read_initramfs_config()

        # Even if initramfs data is not provided, we still try to
        # configure secondary NICs when network is available through
        # ephemeral DHCP.
        netcfg = initrd_cfg
        if self.ds_cfg.get('configure_secondary_nics', False):
            if self._vnic_metadata != sources.UNSET:
                netcfg = _config_secondary_nics(
                             vnic_metadata=self._vnic_metadata,
                             initrd_cfg=initrd_cfg,
                             mtu=self.ds_cfg.get('mtu', DEFAULT_MTU))
            elif initrd_cfg is not None:
                LOG.warning('Unexpected call to network_config when '
                    '_vnic_metadata is None.')

        self._network_config = netcfg if netcfg \
            else self.distro.generate_fallback_config()

        return self._network_config


def _read_system_uuid():
    sys_uuid = util.read_dmi_data('system-uuid')
    return None if sys_uuid is None else sys_uuid.lower()


def _is_platform_viable():
    asset_tag = util.read_dmi_data('chassis-asset-tag')
    return asset_tag == CHASSIS_ASSET_TAG


def _is_iscsi_root():
    return bool(cmdline.read_initramfs_config())


def _load_index(content):
    """Return a list entries parsed from content.

    OpenStack's metadata service returns a newline delimited list
    of items.  Oracle's implementation has html formatted list of links.
    The parser here just grabs targets from <a href="target">
    and throws away "../".

    Oracle has accepted that to be buggy and may fix in the future
    to instead return a '\n' delimited plain text list.  This function
    will continue to work if that change is made."""
    if not content.lower().startswith("<html>"):
        return content.splitlines()
    items = re.findall(
        r'href="(?P<target>[^"]*)"', content, re.MULTILINE | re.IGNORECASE)
    return [i for i in items if not i.startswith(".")]


def _config_secondary_nics(vnic_metadata, initrd_cfg, mtu):
    return _parse_network_config(vnic_metadata, initrd_cfg, mtu)

def parse_network_config(vnic_metadata):
    return _parse_network_config(vnic_metadata)

def _parse_network_config(vnic_metadata, initrd_cfg=None,
                          default_mtu=DEFAULT_MTU):
    """Convert vnic_metadata dictionary to network v1 configuration.

    Parses network configuration from vnic metadata service if present

    @param: vnic_metadata: Dict of content read from vnic metadata service.
    @param: initrd_cfg: initramfs inferred config if available
    @return: Dictionary containing network version 1 standard configuration.
    """

    def vnic_ifname_by_macaddr(macaddr, ifaces):

        if macaddr is None:
            LOG.debug('Oracle: attribute "macAddr" missing in vnic metadata')
            return None

        macaddr = macaddr.lower()

        if macaddr not in ifaces:
            LOG.debug('Oracle: no matching interface found for macAddr=%s', macaddr)
            return None

        return ifaces[macaddr]

    if vnic_metadata and vnic_metadata != sources.UNSET:
        # Assumption: initrd_cfg should actually be initramfs provided config
        # data which is represented using v1 network config format. If in future
        # time the initramfs config is represented using other format than v1,
        # implementation here should be re-adjusted accordingly.
        netconfig = {'version': 1, 'config': []} if not initrd_cfg else initrd_cfg
        ifaces = get_interfaces_by_mac()
        primary = True

        for vnic in vnic_metadata:
            dev_config = {}
            vifname = vnic_ifname_by_macaddr(vnic['macAddr'], ifaces)
            if vifname:
                dev_config['type'] = 'physical'
                dev_config['name'] = vifname
                dev_config['mac_address'] = vnic['macAddr'].lower()
                # Primary VNIC should always get configured through DHCP.
                if primary:
                    dev_config['subnets'] = [{'type' : 'dhcp'}]
                # DHCP service is not running on secondary VNIC.
                # Use static configuration instead.
                else:
                    dev_config['mtu'] = default_mtu
                    privateIp = vnic['privateIp']
                    netPrefix = vnic['subnetCidrBlock'].split('/')[1]
                    dev_config['subnets'] = [{'type' : 'static',
                        'address' : '{ip}/{prefix}'.format(
                            ip=privateIp, prefix=netPrefix)
                        }]
                # If the primary VNIC already gets configured via
                # initramfs, don't generate network config for
                # the netdev. Secondary VNIC network configs will
                # be merged with the initramfs provided initrd_cfg.
                if not primary or not initrd_cfg:
                    netconfig['config'].append(dev_config)

            # nicIndex means Bare Metal Instance, while we only care VM for now
            if 'nicIndex' in vnic:
                LOG.warning('Oracle: secondary VNIC configuration on bare metal instance '
                            'not supported currently.')
                break
            elif not primary and bool(dev_config):
                LOG.debug('Oracle: generating secondary VNIC config from opc v1 metadata: %s',
                          dev_config)

            primary = False

    return netconfig


def read_metadata(endpoint_base=METADATA_ENDPOINT, sys_uuid=None,
                  version='2013-10-17'):
    """Read metadata, return a dictionary.

    Each path listed in the index will be represented in the dictionary.
    If the path ends in .json, then the content will be decoded and
    populated into the dictionary.

    The system uuid (/sys/class/dmi/id/product_uuid) is also populated.
    Example: given paths = ('user_data', 'meta_data.json')
    This would return:
      {version: {'user_data': b'blob', 'meta_data': json.loads(blob.decode())
                 'system_uuid': '3b54f2e0-3ab2-458d-b770-af9926eee3b2'}}
    """
    endpoint = combine_url(endpoint_base, version) + "/"
    if sys_uuid is None:
        sys_uuid = _read_system_uuid()
    if not sys_uuid:
        raise sources.BrokenMetadata("Failed to read system uuid.")

    try:
        resp = readurl(endpoint)
        if not resp.ok():
            raise sources.BrokenMetadata(
                "Bad response from %s: %s" % (endpoint, resp.code))
    except UrlError as e:
        raise sources.BrokenMetadata(
            "Failed to read index at %s: %s" % (endpoint, e))

    entries = _load_index(resp.contents.decode('utf-8'))
    LOG.debug("index url %s contained: %s", endpoint, entries)

    # meta_data.json is required.
    mdj = 'meta_data.json'
    if mdj not in entries:
        raise sources.BrokenMetadata(
            "Required field '%s' missing in index at %s" % (mdj, endpoint))

    ret = {'system_uuid': sys_uuid}
    for path in entries:
        response = readurl(combine_url(endpoint, path))
        if path.endswith(".json"):
            ret[path.rpartition(".")[0]] = (
                json.loads(response.contents.decode('utf-8')))
        else:
            ret[path] = response.contents

    return {version: ret}


def read_vnic_metadata():
    try:
        resp = readurl(VNIC_METADATA_ENDPOINT)
        if not resp.ok():
            raise sources.BrokenMetadata(
                "Bad response from %s: %s" % (VNIC_METADATA_ENDPOINT, resp.code))
    except UrlError as e:
        raise sources.BrokenMetadata(
            "Failed to get vnic metadata at %s: %s" % (VNIC_METADATA_ENDPOINT, e))

    return json.loads(resp.contents.decode('utf-8'))


# Used to match classes to dependencies
datasources = [
    (DataSourceOracle, (sources.DEP_FILESYSTEM,)),
]


# Return a list of data sources that match this set of dependencies
def get_datasource_list(depends):
    return sources.list_from_depends(depends, datasources)


if __name__ == "__main__":
    import argparse
    import os

    parser = argparse.ArgumentParser(description='Query Oracle Cloud Metadata')
    parser.add_argument("--endpoint", metavar="URL",
                        help="The url of the metadata service.",
                        default=METADATA_ENDPOINT)
    args = parser.parse_args()
    sys_uuid = "uuid-not-available-not-root" if os.geteuid() != 0 else None

    data = read_metadata(endpoint_base=args.endpoint, sys_uuid=sys_uuid)
    data['is_platform_viable'] = _is_platform_viable()
    print(util.json_dumps(data))

# vi: ts=4 expandtab
